@@ -43,7 +43,8 @@ KURYR_CRD_GROUP = 'openstack.org'
 KURYR_CRD_VERSION = 'v1'
 KURYR_NET_CRD_PLURAL = 'kuryrnets'
 KURYR_NETWORK_CRD_PLURAL = 'kuryrnetworks'
-KURYR_NET_POLICY_CRD_PLURAL = 'kuryrnetpolicies'
+KURYR_PORT_CRD_PLURAL = 'kuryrports'
+KURYR_LOAD_BALANCER_CRD_PLURAL = 'kuryrloadbalancers'
 K8S_ANNOTATION_PREFIX = 'openstack.org/kuryr'
 K8S_ANNOTATION_LBAAS_STATE = K8S_ANNOTATION_PREFIX + '-lbaas-state'
 K8S_ANNOTATION_LBAAS_RT_STATE = K8S_ANNOTATION_PREFIX + '-lbaas-route-state'
@@ -92,7 +93,9 @@ class BaseKuryrScenarioTest(manager.NetworkScenarioTest):
                               ingress_ipblock_except=[],
                               egress_port=None, egress_port_protocol='TCP',
                               egress_ipblock_cidr=None,
-                              egress_ipblock_except=[]):
+                              egress_ipblock_except=[],
+                              ingress_match_expressions=None,
+                              egress_match_expressions=None):
         if not name:
             name = data_utils.rand_name(prefix='kuryr-network-policy')
         np = k8s_client.V1NetworkPolicy()
@@ -100,15 +103,15 @@ class BaseKuryrScenarioTest(manager.NetworkScenarioTest):
         np.api_version = 'networking.k8s.io/v1'
         np.metadata = k8s_client.V1ObjectMeta(name=name,
                                               namespace=namespace)
-        to, _from = None, None
+        to, _from = [], []
         if egress_ipblock_cidr:
-            to = [k8s_client.V1NetworkPolicyPeer(
+            to.append(k8s_client.V1NetworkPolicyPeer(
                 ip_block=k8s_client.V1IPBlock(cidr=egress_ipblock_cidr,
-                                              _except=egress_ipblock_except))]
+                                              _except=egress_ipblock_except)))
         if ingress_ipblock_cidr:
-            _from = [k8s_client.V1NetworkPolicyPeer(
+            _from.append(k8s_client.V1NetworkPolicyPeer(
                 ip_block=k8s_client.V1IPBlock(cidr=ingress_ipblock_cidr,
-                                              _except=ingress_ipblock_except))]
+                                              _except=ingress_ipblock_except)))
         if ingress_port:
             ingress_port = [k8s_client.V1NetworkPolicyPort(
                 port=ingress_port, protocol=ingress_port_protocol)]
@@ -116,6 +119,14 @@ class BaseKuryrScenarioTest(manager.NetworkScenarioTest):
             egress_port = [k8s_client.V1NetworkPolicyPort(
                 port=egress_port, protocol=egress_port_protocol)]
 
+        if ingress_match_expressions:
+            _from.append(k8s_client.V1NetworkPolicyPeer(
+                pod_selector=k8s_client.V1LabelSelector(
+                    match_expressions=ingress_match_expressions)))
+        if egress_match_expressions:
+            to.append(k8s_client.V1NetworkPolicyPeer(
+                pod_selector=k8s_client.V1LabelSelector(
+                    match_expressions=egress_match_expressions)))
         np.spec = k8s_client.V1NetworkPolicySpec(
             egress=[k8s_client.V1NetworkPolicyEgressRule(
                 ports=egress_port,
@@ -155,7 +166,7 @@ class BaseKuryrScenarioTest(manager.NetworkScenarioTest):
                 name=np_name, namespace=np_namespace)
 
     @classmethod
-    def create_pod(cls, name=None, labels=None, image='kuryr/demo',
+    def create_pod(cls, name=None, labels=None, image='quay.io/kuryr/demo',
                    namespace="default", annotations=None, wait_for_status=True,
                    affinity=None):
         if not name:
@@ -272,19 +283,44 @@ class BaseKuryrScenarioTest(manager.NetworkScenarioTest):
                                        container_name=container_name)
 
     def get_pod_port(self, pod_name, namespace="default"):
-        pod = self.k8s_client.CoreV1Api().read_namespaced_pod_status(
-            namespace=namespace, name=pod_name)
-        kuryr_if = json.loads(pod.metadata.annotations[
-            'openstack.org/kuryr-vif'])
+        try:
+            # TODO(gryf): Current approach is to look at the KuryrPort CRD,
+            # and if it doesn't exists, fallback to check the pod annotations
+            crd = (self.k8s_client.CustomObjectsApi()
+                   .get_namespaced_custom_object(group=KURYR_CRD_GROUP,
+                                                 version=KURYR_CRD_VERSION,
+                                                 namespace=namespace,
+                                                 plural=KURYR_PORT_CRD_PLURAL,
+                                                 name=pod_name))
+            try:
+                vif = [v['vif'].get('versioned_object.data', {}).get('id')
+                       for k, v in crd['status']['vifs'].items()
+                       if v.get('default')]
+            except KeyError:
+                # TODO(gryf): Remove this after moving vifs to status succeed.
+                vif = [v['vif'].get('versioned_object.data', {}).get('id')
+                       for k, v in crd['spec']['vifs'].items()
+                       if v.get('default')]
 
-        # FIXME(dulek): We need this compatibility code to run stable/queens.
-        #               Remove this once it's no longer supported.
-        if 'eth0' in kuryr_if:
-            kuryr_if = kuryr_if['eth0']
-        elif kuryr_if.get('versioned_object.name') == 'PodState':
-            kuryr_if = kuryr_if['versioned_object.data']['default_vif']
+            if vif and vif[0]:
+                return vif[0]
+            else:
+                return None
+        except kubernetes.client.rest.ApiException:
+            pod = self.k8s_client.CoreV1Api().read_namespaced_pod_status(
+                namespace=namespace, name=pod_name)
+            kuryr_if = json.loads(pod.metadata.annotations[
+                'openstack.org/kuryr-vif'])
 
-        return kuryr_if['versioned_object.data']['id']
+            # FIXME(dulek): We need this compatibility code to run
+            #               stable/queens. Remove this once it's no longer
+            #               supported.
+            if 'eth0' in kuryr_if:
+                kuryr_if = kuryr_if['eth0']
+            elif kuryr_if.get('versioned_object.name') == 'PodState':
+                kuryr_if = kuryr_if['versioned_object.data']['default_vif']
+
+            return kuryr_if['versioned_object.data']['id']
 
     @classmethod
     def get_pod_node_name(cls, pod_name, namespace="default"):
@@ -297,7 +333,7 @@ class BaseKuryrScenarioTest(manager.NetworkScenarioTest):
 
     def exec_command_in_pod(self, pod_name, command, namespace="default",
                             stderr=False, container=None,
-                            req_timeout=10, f_timeout=2):
+                            req_timeout=60, f_timeout=2):
         api = self.k8s_client.CoreV1Api()
         kwargs = dict(command=command, stdin=False, stdout=True, tty=False,
                       stderr=stderr)
@@ -319,7 +355,7 @@ class BaseKuryrScenarioTest(manager.NetworkScenarioTest):
                     kwargs['_request_timeout'] = req_timeout
                 if stderr:
                     kwargs['_preload_content'] = False
-                    resp = stream(api.connect_get_namespaced_pod_exec,
+                    resp = stream(api.connect_post_namespaced_pod_exec,
                                   pod_name, namespace, **kwargs)
                     # Run until completion
                     resp.run_forever(timeout=f_timeout)
@@ -393,45 +429,89 @@ class BaseKuryrScenarioTest(manager.NetworkScenarioTest):
     def get_service_ip(
             cls, service_name, spec_type="ClusterIP", namespace="default"):
         api = cls.k8s_client.CoreV1Api()
-        while True:
+        service = api.read_namespaced_service(service_name, namespace)
+        if spec_type == "LoadBalancer":
+            # In case of a cloud provider not being configured, OpenShift
+            # allocates an external IP and overwrites the service
+            # status/ingress/IP set by Kuryr-controller.
+            # In this case, we should retrieve the external IP from
+            # Kuryr's annotation.
+            if CONF.kuryr_kubernetes.kuryrloadbalancers:
+                return cls.get_svc_ip_on_crd(service_name, namespace)
+            else:
+                return cls.get_svc_ip_on_annotation(service_name, namespace)
+        elif spec_type == "ClusterIP":
+            return service.spec.cluster_ip
+        else:
+            raise lib_exc.NotImplemented()
+
+    @classmethod
+    def get_svc_ip_on_annotation(cls, service_name, namespace):
+        api = cls.k8s_client.CoreV1Api()
+        start = time.time()
+        while time.time() - start < consts.LB_TIMEOUT:
             time.sleep(5)
             service = api.read_namespaced_service(service_name, namespace)
-            if spec_type == "LoadBalancer":
-                # In case of a cloud provider not being configured, OpenShift
-                # allocates an external IP and overwrites the service
-                # status/ingress/IP set by Kuryr-controller.
-                # In this case, we should retrieve the external IP from
-                # Kuryr's annotation.
+            if service.status.load_balancer.ingress:
+                endpoints = api.read_namespaced_endpoints(
+                    service_name, namespace)
+                annotations = endpoints.metadata.annotations
+                try:
+                    ann_dict = json.loads(
+                        annotations[K8S_ANNOTATION_LBAAS_STATE])
+                    ann_lb_ip = (
+                        ann_dict["versioned_object.data"]
+                        ["service_pub_ip_info"]
+                        ["versioned_object.data"]
+                        ["ip_addr"])
+                except KeyError:
+                    LOG.info("Waiting till LB's IP appears in annotation "
+                             "(ingress.ip=%s)",
+                             service.status.load_balancer.ingress[0].ip)
+                    continue
+                if ann_lb_ip != service.status.load_balancer.ingress[0].ip:
+                    LOG.warning(
+                        'Annotated pub_ip(%s) != ingress.ip(%s).',
+                        ann_lb_ip,
+                        service.status.load_balancer.ingress[0].ip)
+                    if not CONF.kuryr_kubernetes.cloud_provider:
+                        return ann_lb_ip
+                return service.status.load_balancer.ingress[0].ip
+        msg = "Timed out waiting for endpoints annotation %s" % service_name
+        raise lib_exc.TimeoutException(msg)
 
-                if service.status.load_balancer.ingress:
-                    endpoints = api.read_namespaced_endpoints(
+    @classmethod
+    def get_svc_ip_on_crd(cls, service_name, namespace):
+        api = cls.k8s_client.CoreV1Api()
+        start = time.time()
+        while time.time() - start < consts.LB_TIMEOUT:
+            time.sleep(5)
+            service = api.read_namespaced_service(service_name, namespace)
+            if service.status.load_balancer.ingress:
+                try:
+                    klb_crd = cls.get_kuryr_loadbalancer_crds(
                         service_name, namespace)
-                    annotations = endpoints.metadata.annotations
+                except kubernetes.client.rest.ApiException:
+                    continue
+                ingress_ip = service.status.load_balancer.ingress[0].ip
+                klb_status = klb_crd.get('status')
+                if (klb_status and klb_status.get('service_pub_ip_info')):
                     try:
-                        ann_dict = json.loads(
-                            annotations[K8S_ANNOTATION_LBAAS_STATE])
-                        ann_lb_ip = (
-                            ann_dict["versioned_object.data"]
-                            ["service_pub_ip_info"]
-                            ["versioned_object.data"]
-                            ["ip_addr"])
+                        crd_lb_ip = klb_status['service_pub_ip_info'][
+                            'ip_addr']
                     except KeyError:
-                        LOG.info("Waiting till LB's IP appears in annotation "
-                                 "(ingress.ip=%s)",
-                                 service.status.load_balancer.ingress[0].ip)
+                        LOG.info("Waiting till LB's IP appears in CRD "
+                                 "(ingress.ip=%s)", ingress_ip)
                         continue
-                    if ann_lb_ip != service.status.load_balancer.ingress[0].ip:
-                        LOG.warning(
-                            'Annotated pub_ip(%s) != ingress.ip(%s).',
-                            ann_lb_ip,
-                            service.status.load_balancer.ingress[0].ip)
-                        if not CONF.kuryr_kubernetes.cloud_provider:
-                            return ann_lb_ip
-                    return service.status.load_balancer.ingress[0].ip
-            elif spec_type == "ClusterIP":
-                return service.spec.cluster_ip
-            else:
-                raise lib_exc.NotImplemented()
+                if crd_lb_ip != ingress_ip:
+                    LOG.warning(
+                        'LB CRD pub_ip(%s) != ingress.ip(%s).',
+                        crd_lb_ip, ingress_ip)
+                    if not CONF.kuryr_kubernetes.cloud_provider:
+                        return crd_lb_ip
+                return ingress_ip
+        msg = "Timed out waiting for lb crd status %s" % service_name
+        raise lib_exc.TimeoutException(msg)
 
     @classmethod
     def wait_kuryr_annotation(cls, group, version, plural, annotation,
@@ -575,7 +655,77 @@ class BaseKuryrScenarioTest(manager.NetworkScenarioTest):
                                         namespace=namespace)
 
     @classmethod
-    def create_namespace(cls, name=None, wait_for_crd=True):
+    def service_without_selector_base(cls, pod_num=2,
+                                      spec_type="ClusterIP",
+                                      protocol="TCP",
+                                      port=80,
+                                      target_port=8080,
+                                      namespace="default",
+                                      get_ip=True,
+                                      service_name=None):
+
+        """Create a setup for services without selector
+
+        This method creates a pod, service without selector and endpoint.
+        """
+        pod_name_list = []
+        for i in range(pod_num):
+            pod_name, pod = cls.create_pod(namespace=namespace)
+            pod_name_list.append(pod_name)
+
+        cls.pod_num = pod_num
+        service_name, service_obj = cls.create_service(
+            spec_type=spec_type, protocol=protocol, port=port,
+            target_port=target_port, namespace=namespace,
+            service_name=service_name, pod_label=None)
+        pod_1_ip = cls.get_pod_ip(pod_name_list[0], namespace=namespace)
+        pod_2_ip = cls.get_pod_ip(pod_name_list[1], namespace=namespace)
+        port = service_obj.spec.ports[0]
+        port_name = port.name
+        # Endpoints is a collection of endpoints that implement
+        # the actual service
+        endpoint = cls.k8s_client.V1Endpoints()
+        endpoint.metadata = cls.k8s_client.V1ObjectMeta(name=service_name)
+        # EndpointSubset is a group of addresses with a set of ports
+        endpoint.subsets = [cls.k8s_client.V1EndpointSubset(
+                            addresses=[
+                                cls.k8s_client.V1EndpointAddress(
+                                    ip=pod_1_ip),
+                                cls.k8s_client.V1EndpointAddress(
+                                    ip=pod_2_ip
+                                )
+                            ],
+                            ports=[cls.k8s_client.V1EndpointPort(
+                                  name=port_name,
+                                  port=target_port,
+                                  protocol=protocol)])]
+        cls.k8s_client.CoreV1Api().create_namespaced_endpoints(
+            namespace=namespace, body=endpoint)
+        if get_ip:
+            cls.service_ip = cls.get_service_ip(
+                service_name, spec_type=spec_type, namespace=namespace)
+            cls.verify_lbaas_endpoints_configured(service_name, pod_num,
+                                                  namespace)
+            cls.service_name = service_name
+
+    @classmethod
+    def wait_for_ns_serviceaccount(cls, namespace,
+                                   timeout_period=consts.NS_TIMEOUT):
+        start = time.time()
+        while time.time() - start < timeout_period:
+            time.sleep(5)
+            try:
+                cls.k8s_client.CoreV1Api().read_namespaced_service_account(
+                    'default', namespace)
+                return True
+            except kubernetes.client.rest.ApiException:
+                pass
+
+        return False
+
+    @classmethod
+    def create_namespace(cls, name=None, wait_for_crd=True,
+                         timeout_period=consts.NS_TIMEOUT):
         if not name:
             name = data_utils.rand_name(prefix='kuryr-namespace')
         namespace = cls.k8s_client.V1Namespace()
@@ -583,13 +733,22 @@ class BaseKuryrScenarioTest(manager.NetworkScenarioTest):
         namespace_obj = cls.k8s_client.CoreV1Api().create_namespace(
             body=namespace)
 
-        if not wait_for_crd:
+        if not wait_for_crd or not CONF.kuryr_kubernetes.subnet_per_namespace:
+            # You cannot create pods until default ServiceAccount for the
+            # namespace is created, let's wait for it.
+            if not cls.wait_for_ns_serviceaccount(name):
+                raise lib_exc.TimeoutException(
+                    "Timed out waiting for default ServiceAccount to get "
+                    "created in %s namespace. Is "
+                    "kubernetes-controller-manager running?" % name)
+
             return name, namespace_obj
 
+        start = time.time()
         if CONF.kuryr_kubernetes.kuryrnetworks:
             # wait until kuryrnetwork CRD gets populated
-            while True:
-                time.sleep(1)
+            while time.time() - start < timeout_period:
+                time.sleep(5)
                 try:
                     kns_crd = cls.get_kuryr_network_crds(name)
                 except kubernetes.client.rest.ApiException:
@@ -597,6 +756,10 @@ class BaseKuryrScenarioTest(manager.NetworkScenarioTest):
                 if (kns_crd.get('status') and
                         kns_crd['status'].get('routerId')):
                     break
+            if (not kns_crd.get('status')) or (
+                    not kns_crd.get('status').get('routerId')):
+                msg = 'Timed out waiting for kns crd %s creation' % name
+                raise lib_exc.TimeoutException(msg)
         else:
             kuryr_crd_annotation = K8S_ANNOTATION_PREFIX + "-net-crd"
             # wait until namespace gets created
@@ -656,11 +819,11 @@ class BaseKuryrScenarioTest(manager.NetworkScenarioTest):
             name=namespace)
 
     @classmethod
-    def get_kuryr_netpolicy_crds(cls, name, namespace='default', **kwargs):
+    def get_kuryr_loadbalancer_crds(cls, name, namespace):
         return cls.k8s_client.CustomObjectsApi().get_namespaced_custom_object(
             group=KURYR_CRD_GROUP, version=KURYR_CRD_VERSION,
-            namespace=namespace, plural=KURYR_NET_POLICY_CRD_PLURAL,
-            name=name, **kwargs)
+            namespace=namespace, plural=KURYR_LOAD_BALANCER_CRD_PLURAL,
+            name=name)
 
     def get_pod_list(self, namespace='default', label_selector=''):
         return self.k8s_client.CoreV1Api().list_namespaced_pod(
@@ -670,10 +833,18 @@ class BaseKuryrScenarioTest(manager.NetworkScenarioTest):
         return [pod.metadata.name for pod in self.get_pod_list(
             namespace=namespace, label_selector=label_selector)]
 
+    def get_pod_ip_list(self, namespace='default', label_selector=''):
+        return [pod.status.pod_ip for pod in self.get_pod_list(
+            namespace=namespace, label_selector=label_selector)]
+
     def get_controller_pod_names(self):
-        return self.get_pod_name_list(
+        controller_label = CONF.kuryr_kubernetes.controller_label
+        controller_pod_names = self.get_pod_name_list(
             namespace=CONF.kuryr_kubernetes.kube_system_namespace,
-            label_selector=CONF.kuryr_kubernetes.controller_label)
+            label_selector=controller_label)
+        self.assertNotEmpty(controller_pod_names, "Can't find controller pods "
+                            "with label %s" % controller_label)
+        return controller_pod_names
 
     def _run_and_assert_fn(self, fn, repeats=10, responses_num=2):
         cmd_outputs = set()
@@ -810,18 +981,52 @@ class BaseKuryrScenarioTest(manager.NetworkScenarioTest):
     @classmethod
     def verify_lbaas_endpoints_configured(cls, ep_name, pod_num,
                                           namespace='default'):
-        cls._verify_endpoints_annotation(
-            ep_name=ep_name, ann_string=K8S_ANNOTATION_LBAAS_STATE,
-            poll_interval=5, namespace=namespace, pod_num=pod_num)
+        if CONF.kuryr_kubernetes.kuryrloadbalancers:
+            cls._verify_klb_crd(
+                ep_name, poll_interval=5, namespace=namespace,
+                pod_num=pod_num)
+        else:
+            cls._verify_endpoints_annotation(
+                ep_name=ep_name, ann_string=K8S_ANNOTATION_LBAAS_STATE,
+                poll_interval=5, namespace=namespace, pod_num=pod_num)
+
+    @classmethod
+    def _verify_klb_crd(cls, name, poll_interval=1, namespace='default',
+                        timeout_period=consts.LB_TIMEOUT, pod_num=None):
+        start = time.time()
+        klb_crd_has_status = False
+        while time.time() - start < timeout_period:
+            time.sleep(poll_interval)
+            try:
+                klb_crd = cls.get_kuryr_loadbalancer_crds(
+                    name, namespace)
+            except kubernetes.client.rest.ApiException:
+                continue
+            if (klb_crd.get('status') and
+                    klb_crd['status'].get('members')):
+                members_num = len(klb_crd['status'].get('members'))
+                if pod_num and pod_num != members_num:
+                    LOG.info("members num(%d) != pod_num(%d)",
+                             members_num, pod_num)
+                    continue
+                klb_crd_has_status = True
+                break
+        if not klb_crd_has_status:
+            msg = "Timed out waiting for klb crd %s creation" % name
+            raise lib_exc.TimeoutException(msg)
 
     @classmethod
     def _verify_endpoints_annotation(cls, ep_name, ann_string,
                                      poll_interval=1, namespace='default',
-                                     pod_num=None):
+                                     pod_num=None,
+                                     timeout_period=consts.LB_TIMEOUT):
         LOG.info("Look for %s string in ep=%s annotation ",
                  ann_string, ep_name)
+
+        start = time.time()
+        annotation_success = False
         # wait until endpoint annotation created
-        while True:
+        while time.time() - start < timeout_period:
             time.sleep(poll_interval)
             ep = cls.k8s_client.CoreV1Api().read_namespaced_endpoints(
                 ep_name, namespace)
@@ -846,11 +1051,16 @@ class BaseKuryrScenarioTest(manager.NetworkScenarioTest):
                         continue
                 LOG.info("Found %s string in ep=%s annotation ",
                          ann_string, ep_name)
-                return
+                annotation_success = True
+                break
             except KeyError:
-                LOG.info("Waiting till %s will appears "
+                LOG.info("Waiting till %s will appear "
                          "in ep=%s annotation ", ann_string, ep_name)
                 continue
+        if not annotation_success:
+            msg = "Timed out waiting for %s in ep=%s annotation to appear" % (
+                    ann_string, ep_name)
+            raise lib_exc.TimeoutException(msg)
 
     def create_vm_for_connectivity_test(self):
         keypair = self.create_keypair()
@@ -1081,21 +1291,43 @@ class BaseKuryrScenarioTest(manager.NetworkScenarioTest):
     def check_service_internal_connectivity(self, service_port='80',
                                             protocol='TCP',
                                             namespace='default',
+                                            labels=None,
+                                            pod_num=None,
+                                            pod_name=None,
                                             cleanup=True):
+        """Verify client pod to service connectivity
+
+        Create a pod unless a value for the pod_name parameter is provided and
+        check connectivity to a service from that pod.
+
+        :param service_port - The port of the service we check
+        :param protocol - The service protocol we check
+        :namespace - The namespace of the client pod
+        :param labels - The labels of the client pod
+        :param pod_num - The number of pods expected to serve the service
+        :param pod_name - If supplied no pod will be created and instead a pod
+                          with this name will be used
+        :param cleanup - Whether to add a cleanup function for the created pod
+        :returns: The name of the client pod that was created or passed to
+                  the function
+        """
         # FIXME(itzikb): Use the clusterIP to
         # check service status as there are some issues with the FIPs
         # and OVN gates
         clusterip_svc_ip = self.get_service_ip(self.service_name,
                                                spec_type='ClusterIP',
                                                namespace=namespace)
-        pod_name, pod = self.create_pod(namespace=namespace)
-        if cleanup:
-            self.addClassResourceCleanup(self.delete_pod, pod_name,
-                                         namespace=namespace)
+        pod_num = pod_num or self.pod_num
+        if not pod_name:
+            pod_name, _ = self.create_pod(namespace=namespace, labels=labels)
+            if cleanup:
+                self.addClassResourceCleanup(self.delete_pod, pod_name,
+                                             namespace=namespace)
         self.assert_backend_amount_from_pod(
             clusterip_svc_ip,
-            self.pod_num,
+            pod_num,
             pod_name,
             service_port,
             protocol,
             namespace_name=namespace)
+        return pod_name
