@@ -19,3 +19,5 @@ POD_AFFINITY = {'requiredDuringSchedulingIgnoredDuringExecution': [
         'topologyKey': 'kubernetes.io/hostname'}]}
 TIME_TO_APPLY_SGS = 30
 POD_STATUS_RETRIES = 240
+NS_TIMEOUT = 600
+LB_TIMEOUT = 1200
